import time
from typing import NamedTuple

import pytest
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, padding, rsa, x25519

from yubikit.core import TRANSPORT, InvalidPinError
from yubikit.core.smartcard import AID, ApduError
from yubikit.management import CAPABILITY
from yubikit.openpgp import (
    KEY_REF,
    OID,
    RSA_SIZE,
    KdfIterSaltedS2k,
    KdfNone,
    OpenPgpSession,
)

from . import condition

E = 65537
DEFAULT_PIN = "123456"
NON_DEFAULT_PIN = "12345670"
DEFAULT_ADMIN_PIN = "12345678"
NON_DEFAULT_ADMIN_PIN = "12345670"


@pytest.fixture
@condition.capability(CAPABILITY.OPENPGP)
def session(ccid_connection, info, scp_params):
    if ccid_connection.transport == TRANSPORT.NFC and fips_capable(info):
        pgp = OpenPgpSession(ccid_connection, scp_params)
    else:
        pgp = OpenPgpSession(ccid_connection)
    pgp.reset()
    return pgp


class Keys(NamedTuple):
    pin: str
    admin: str


def not_roca(version):
    """ROCA affected"""
    return not ((4, 2, 0) <= version < (4, 3, 5))


def fips_capable(info):
    """Not FIPS capable"""
    return CAPABILITY.OPENPGP in info.fips_capable


def not_fips_capable(info):
    """FIPS capable"""
    return not fips_capable(info)


@pytest.fixture
def keys(session, info, transport, scp_params):
    if fips_capable(info) or info.pin_complexity:
        new_keys = Keys(
            "12345679",
            "12345679",
        )
        session.change_pin(DEFAULT_PIN, new_keys.pin)
        session.change_admin(DEFAULT_ADMIN_PIN, new_keys.admin)

        session.protocol.connection.connection.disconnect()
        session.protocol.connection.connection.connect()
        session.protocol.select(AID.OPENPGP)
        if transport == TRANSPORT.NFC and scp_params:
            session.protocol.init_scp(scp_params)

        yield new_keys
    else:
        yield Keys(DEFAULT_PIN, DEFAULT_ADMIN_PIN)


def test_change_pin(session, keys):
    session.verify_pin(keys.pin)
    session.change_pin(keys.pin, NON_DEFAULT_PIN)

    with pytest.raises(InvalidPinError) as e:
        session.verify_pin(keys.pin)
        assert e.value.attempts_remaining == 2

    session.verify_pin(NON_DEFAULT_PIN)
    session.change_pin(NON_DEFAULT_PIN, keys.pin)

    with pytest.raises(InvalidPinError) as e:
        session.verify_pin(NON_DEFAULT_PIN)
        assert e.value.attempts_remaining == 2

    session.verify_pin(keys.pin)


def test_change_admin(session, keys):
    session.verify_admin(keys.admin)
    session.change_admin(keys.admin, NON_DEFAULT_ADMIN_PIN)

    with pytest.raises(InvalidPinError) as e:
        session.verify_admin(keys.admin)
        assert e.value.attempts_remaining == 2

    session.verify_admin(NON_DEFAULT_ADMIN_PIN)
    session.change_admin(NON_DEFAULT_ADMIN_PIN, keys.admin)

    with pytest.raises(InvalidPinError) as e:
        session.verify_admin(NON_DEFAULT_ADMIN_PIN)
        assert e.value.attempts_remaining == 2

    session.verify_admin(keys.admin)


def test_change_pin_retries(session, keys, version):
    with pytest.raises(ApduError):
        session.set_pin_attempts(5, 0, 25)

    session.verify_admin(keys.admin)
    session.set_pin_attempts(5, 0, 25)

    assert session.get_pin_status().attempts_user == 5
    if version >= (4, 0, 0):
        # NEO returns 3 if no reset code is set
        assert session.get_pin_status().attempts_reset == 0
    assert session.get_pin_status().attempts_admin == 25

    with pytest.raises(InvalidPinError) as e:
        session.verify_pin(NON_DEFAULT_PIN)
        assert e.value.attempts_remaining == 4

    with pytest.raises(InvalidPinError) as e:
        session.verify_admin(NON_DEFAULT_ADMIN_PIN)
        assert e.value.attempts_remaining == 24


def test_import_requires_admin(session):
    priv = rsa.generate_private_key(E, RSA_SIZE.RSA2048, default_backend())
    with pytest.raises(ApduError):
        session.put_key(KEY_REF.SIG, priv)


@condition.check(not_roca)
def test_generate_requires_admin(session):
    with pytest.raises(ApduError):
        session.generate_rsa_key(KEY_REF.SIG, RSA_SIZE.RSA2048)


@condition.min_version(5, 2)
@pytest.mark.parametrize("oid", [x for x in OID if "25519" not in x.name])
def test_import_sign_ecdsa(session, info, keys, oid):
    if fips_capable(info) and oid == OID.SECP256K1:
        pytest.skip("FIPS capable")

    priv = ec.generate_private_key(getattr(ec, oid.name)())
    session.verify_admin(keys.admin)
    session.put_key(KEY_REF.SIG, priv)
    message = b"Hello world"
    session.verify_pin(keys.pin)
    sig = session.sign(message, hashes.SHA256())
    priv.public_key().verify(sig, message, ec.ECDSA(hashes.SHA256()))


@condition.min_version(5, 2)
def test_import_sign_eddsa(session, keys):
    priv = ed25519.Ed25519PrivateKey.generate()
    session.verify_admin(keys.admin)
    session.put_key(KEY_REF.SIG, priv)
    message = b"Hello world"
    session.verify_pin(keys.pin)
    sig = session.sign(message, hashes.SHA256())
    priv.public_key().verify(sig, message)


@condition.min_version(5, 2)
@pytest.mark.parametrize("oid", [x for x in OID if "25519" not in x.name])
def test_import_ecdh(session, info, keys, oid):
    if fips_capable(info) and oid == OID.SECP256K1:
        pytest.skip("FIPS capable")
    priv = ec.generate_private_key(getattr(ec, oid.name)())
    session.verify_admin(keys.admin)
    session.put_key(KEY_REF.DEC, priv)
    e_priv = ec.generate_private_key(getattr(ec, oid.name)())
    shared1 = e_priv.exchange(ec.ECDH(), priv.public_key())
    session.verify_pin(keys.pin, extended=True)
    shared2 = session.decrypt(e_priv.public_key())

    assert shared1 == shared2


@condition.check(not_fips_capable)
@condition.min_version(5, 2)
def test_import_ecdh_x25519(session, keys):
    priv = x25519.X25519PrivateKey.generate()
    session.verify_admin(keys.admin)
    session.put_key(KEY_REF.DEC, priv)
    e_priv = x25519.X25519PrivateKey.generate()
    shared1 = e_priv.exchange(priv.public_key())
    session.verify_pin(keys.pin, extended=True)
    shared2 = session.decrypt(e_priv.public_key())

    assert shared1 == shared2


@pytest.mark.parametrize("key_size", [2048, 3072, 4096])
def test_import_sign_rsa(session, keys, key_size, info):
    if key_size != 2048:
        if info.version[0] < 4:
            pytest.skip(f"RSA {key_size} requires YubiKey 4 or later")
        elif info.version[0] == 4 and info.is_fips:
            pytest.skip(f"RSA {key_size} not supported on YubiKey 4 FIPS")
    priv = rsa.generate_private_key(E, key_size, default_backend())
    session.verify_admin(keys.admin)
    session.put_key(KEY_REF.SIG, priv)
    if 0 < info.version[0] < 5:
        # Keys don't work without a generation time (or fingerprint)
        session.set_generation_time(KEY_REF.SIG, int(time.time()))

    message = b"Hello world"
    session.verify_pin(keys.pin)
    sig = session.sign(message, hashes.SHA256())
    priv.public_key().verify(sig, message, padding.PKCS1v15(), hashes.SHA256())


@condition.check(not_fips_capable)
@pytest.mark.parametrize("key_size", [2048, 3072, 4096])
def test_import_decrypt_rsa(session, keys, key_size, info):
    if key_size != 2048:
        if info.version[0] < 4:
            pytest.skip(f"RSA {key_size} requires YubiKey 4 or later")
        elif info.version[0] == 4 and info.is_fips:
            pytest.skip(f"RSA {key_size} not supported on YubiKey 4 FIPS")
    priv = rsa.generate_private_key(E, key_size, default_backend())
    session.verify_admin(keys.admin)
    session.put_key(KEY_REF.DEC, priv)
    if info.version[0] < 5:
        # Keys don't work without a generation time (or fingerprint)
        session.set_generation_time(KEY_REF.DEC, int(time.time()))

    message = b"Hello world"
    cipher = priv.public_key().encrypt(message, padding.PKCS1v15())
    session.verify_pin(keys.pin, extended=True)
    plain = session.decrypt(cipher)

    assert message == plain


@condition.check(not_roca)
@pytest.mark.parametrize("key_size", [2048, 3072, 4096])
def test_generate_rsa(session, keys, key_size, info):
    if key_size != 2048:
        if info.version[0] < 4:
            pytest.skip(f"RSA {key_size} requires YubiKey 4 or later")
        elif info.version[0] == 4 and info.is_fips:
            pytest.skip(f"RSA {key_size} not supported on YubiKey 4 FIPS")
    session.verify_admin(keys.admin)
    pub = session.generate_rsa_key(KEY_REF.SIG, RSA_SIZE(key_size))
    if info.version[0] < 5:
        # Keys don't work without a generation time (or fingerprint)
        session.set_generation_time(KEY_REF.SIG, int(time.time()))

    assert pub.key_size == key_size

    message = b"Hello world"
    session.verify_pin(keys.pin)
    sig = session.sign(message, hashes.SHA256())
    pub.verify(sig, message, padding.PKCS1v15(), hashes.SHA256())


@condition.min_version(5, 2)
@pytest.mark.parametrize("oid", [x for x in OID if "25519" not in x.name])
def test_generate_ecdsa(session, info, keys, oid):
    if fips_capable(info) and oid == OID.SECP256K1:
        pytest.skip("FIPS capable")

    session.verify_admin(keys.admin)
    pub = session.generate_ec_key(KEY_REF.SIG, oid)
    message = b"Hello world"
    session.verify_pin(keys.pin)
    sig = session.sign(message, hashes.SHA256())
    pub.verify(sig, message, ec.ECDSA(hashes.SHA256()))


@condition.min_version(5, 2)
def test_generate_ed25519(session, keys):
    session.verify_admin(keys.admin)
    pub = session.generate_ec_key(KEY_REF.SIG, OID.Ed25519)
    message = b"Hello world"
    session.verify_pin(keys.pin)
    sig = session.sign(message, hashes.SHA256())
    pub.verify(sig, message)


@condition.min_version(5, 2)
@condition.check(not_fips_capable)
def test_generate_x25519(session, keys):
    session.verify_admin(keys.admin)
    pub = session.generate_ec_key(KEY_REF.DEC, OID.X25519)

    e_priv = x25519.X25519PrivateKey.generate()
    shared1 = e_priv.exchange(pub)
    session.verify_pin(keys.pin, extended=True)
    shared2 = session.decrypt(e_priv.public_key())

    assert shared1 == shared2


@condition.min_version(5, 2)
def test_kdf(session, keys):
    with pytest.raises(ApduError):
        session.set_kdf(KdfIterSaltedS2k.create())

    session.change_admin(keys.admin, NON_DEFAULT_ADMIN_PIN)
    session.verify_admin(NON_DEFAULT_ADMIN_PIN)
    session.set_kdf(KdfIterSaltedS2k.create())
    session.verify_admin(DEFAULT_ADMIN_PIN)
    session.verify_pin(DEFAULT_PIN)

    session.change_admin(DEFAULT_ADMIN_PIN, NON_DEFAULT_ADMIN_PIN)
    session.change_pin(DEFAULT_PIN, NON_DEFAULT_PIN)
    session.verify_pin(NON_DEFAULT_PIN)

    session.set_kdf(KdfNone())
    session.verify_admin(DEFAULT_ADMIN_PIN)
    session.verify_pin(DEFAULT_PIN)


@condition.min_version(5, 2)
def test_attestation(session, keys):
    if not session.get_key_information()[KEY_REF.ATT]:
        pytest.skip("No attestation key")

    session.verify_admin(keys.admin)
    pub = session.generate_ec_key(KEY_REF.SIG, OID.SECP256R1)

    session.verify_pin(keys.pin)
    cert = session.attest_key(KEY_REF.SIG)

    assert cert.public_key() == pub


@condition.min_version(5, 2)
def test_get_challenge(session):
    for ln in (1, 4, 8, 100):
        x = session.get_challenge(ln)
        assert len(x) == ln
        if ln > 1:  # Avoid collision too often
            assert x != session.get_challenge(ln)



# Copyright (c) 2020 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import abc
import logging
import struct
import warnings
from enum import Enum, IntEnum, unique
from time import time

from yubikit.logging import LOG_LEVEL

from .. import (
    TRANSPORT,
    USB_INTERFACE,
    ApplicationNotAvailableError,
    BadResponseError,
    CommandError,
    Connection,
    NotSupportedError,
    Version,
)
from .scp import (
    INS_EXTERNAL_AUTHENTICATE,
    Scp03KeyParams,
    Scp11KeyParams,
    ScpKeyParams,
    ScpState,
)

logger = logging.getLogger(__name__)


class SmartCardConnection(Connection, metaclass=abc.ABCMeta):
    usb_interface = USB_INTERFACE.CCID

    @property
    @abc.abstractmethod
    def transport(self) -> TRANSPORT:
        """Get the transport type of the connection (USB or NFC)"""

    @abc.abstractmethod
    def send_and_receive(self, apdu: bytes) -> tuple[bytes, int]:
        """Sends a command APDU and returns the response"""


class ApduError(CommandError):
    """Thrown when an APDU response has the wrong SW code"""

    def __init__(self, data: bytes, sw: int):
        self.data = data
        self.sw = sw

    def __str__(self):
        try:
            name = SW(self.sw).name
            return f"APDU error: SW=0x{self.sw:04x} ({name})"
        except ValueError:
            return f"APDU error: SW=0x{self.sw:04x}"


@unique
class ApduFormat(str, Enum):
    """APDU encoding format"""

    SHORT = "short"
    EXTENDED = "extended"


@unique
class AID(bytes, Enum):
    """YubiKey Application smart card AID values."""

    OTP = bytes.fromhex("a0000005272001")
    MANAGEMENT = bytes.fromhex("a000000527471117")
    OPENPGP = bytes.fromhex("d27600012401")
    OATH = bytes.fromhex("a0000005272101")
    PIV = bytes.fromhex("a000000308")
    FIDO = bytes.fromhex("a0000006472f0001")
    HSMAUTH = bytes.fromhex("a000000527210701")
    SECURE_DOMAIN = bytes.fromhex("a000000151000000")


@unique
class SW(IntEnum):
    NO_INPUT_DATA = 0x6285
    VERIFY_FAIL_NO_RETRY = 0x63C0
    MEMORY_FAILURE = 0x6581
    WRONG_LENGTH = 0x6700
    SECURITY_CONDITION_NOT_SATISFIED = 0x6982
    AUTH_METHOD_BLOCKED = 0x6983
    DATA_INVALID = 0x6984
    CONDITIONS_NOT_SATISFIED = 0x6985
    COMMAND_NOT_ALLOWED = 0x6986
    INCORRECT_PARAMETERS = 0x6A80
    FUNCTION_NOT_SUPPORTED = 0x6A81
    FILE_NOT_FOUND = 0x6A82
    RECORD_NOT_FOUND = 0x6A83
    NO_SPACE = 0x6A84
    REFERENCE_DATA_NOT_FOUND = 0x6A88
    APPLET_SELECT_FAILED = 0x6999
    WRONG_PARAMETERS_P1P2 = 0x6B00
    INVALID_INSTRUCTION = 0x6D00
    CLASS_NOT_SUPPORTED = 0x6E00
    COMMAND_ABORTED = 0x6F00
    OK = 0x9000


INS_SELECT = 0xA4
P1_SELECT = 0x04
P2_SELECT = 0x00

INS_SEND_REMAINING = 0xC0


class ApduFormatter(abc.ABC):
    @abc.abstractmethod
    def format_apdu(
        self, cla: int, ins: int, p1: int, p2: int, data: bytes, le: int
    ) -> bytes: ...


SHORT_APDU_MAX_CHUNK = 0xFF


class ShortApduFormatter(ApduFormatter):
    def format_apdu(self, cla, ins, p1, p2, data, le):
        if len(data) > SHORT_APDU_MAX_CHUNK:
            raise ValueError(
                f"Data length {len(data)} exceeds maximum APDU size "
                f"{SHORT_APDU_MAX_CHUNK}"
            )

        buf = struct.pack(">BBBB", cla, ins, p1, p2)
        if data:
            buf += struct.pack(">B", len(data)) + data
        if le:
            buf += struct.pack(">B", le)
        elif not data:
            # No data nor Le, need explicit Lc
            buf += b"\0"

        return buf


class ExtendedApduFormatter(ApduFormatter):
    def __init__(self, max_apdu_size):
        self._max_apdu_size = max_apdu_size

    def format_apdu(self, cla, ins, p1, p2, data, le):
        buf = struct.pack(">BBBB", cla, ins, p1, p2)
        if data:
            buf += struct.pack(">BH", 0, len(data)) + data
        if le:
            if not data:
                # Use 3-byte Le
                buf += b"\0"
            buf += struct.pack(">H", le)

        if len(buf) > self._max_apdu_size:
            raise NotSupportedError("APDU length exceeds YubiKey capability")
        return buf


class ApduProcessor(abc.ABC):
    @abc.abstractmethod
    def send_apdu(
        self,
        cla: int,
        ins: int,
        p1: int,
        p2: int,
        data: bytes,
        le: int,
    ) -> tuple[bytes, int]: ...


class ApduFormatProcessor(ApduProcessor):
    def __init__(self, connection: SmartCardConnection, formatter: ApduFormatter):
        self.connection = connection
        self.formatter = formatter

    def send_apdu(self, cla, ins, p1, p2, data, le):
        apdu = self.formatter.format_apdu(cla, ins, p1, p2, data, le)
        return self.connection.send_and_receive(apdu)


class CommandChainingProcessor(ApduFormatProcessor):
    def __init__(self, connection: SmartCardConnection):
        super().__init__(connection, ShortApduFormatter())

    def send_apdu(self, cla, ins, p1, p2, data, le):
        while len(data) > SHORT_APDU_MAX_CHUNK:
            chunk, data = (
                data[:SHORT_APDU_MAX_CHUNK],
                data[SHORT_APDU_MAX_CHUNK:],
            )
            response, sw = super().send_apdu(0x10 | cla, ins, p1, p2, chunk, le)
            if sw != SW.OK:
                return response, sw
        return super().send_apdu(cla, ins, p1, p2, data, le)


SW1_HAS_MORE_DATA = 0x61


class ResponseChainingProcessor(ApduProcessor):
    def __init__(
        self, delegate: ApduProcessor, ins_send_remaining: int = INS_SEND_REMAINING
    ):
        self.delegate = delegate
        self.ins_send_remaining = ins_send_remaining

    def send_apdu(self, cla, ins, p1, p2, data, le):
        response, sw = self.delegate.send_apdu(cla, ins, p1, p2, data, le)

        # Read chained response
        buf = b""
        while sw >> 8 == SW1_HAS_MORE_DATA:
            buf += response
            response, sw = self.delegate.send_apdu(
                0, self.ins_send_remaining, 0, 0, b"", 0
            )

        buf += response
        return buf, sw


class TouchWorkaroundProcessor(ApduProcessor):
    def __init__(self, delegate: ApduProcessor):
        self.delegate = delegate
        self._last_long_resp = 0.0

    def send_apdu(self, cla, ins, p1, p2, data, le):
        if self._last_long_resp > 0 and time() - self._last_long_resp < 2:
            logger.debug("Sending dummy APDU as touch workaround")
            # Dummy APDU, returns error
            self.delegate.send_apdu(0, 0, 0, 0, b"", 0)
            self._last_long_resp = 0

        resp, sw = self.delegate.send_apdu(cla, ins, p1, p2, data, le)

        if len(resp) > 54:
            self._last_long_resp = time()
        else:
            self._last_long_resp = 0

        return resp, sw


class _MaxApduSize(IntEnum):
    NEO = 1390
    YK4 = 2038
    YK4_3 = 3062


class ScpProcessor(ApduProcessor):
    # For short APDUs, we need to use the extended processor for MAC calculation
    extended_formatter = ExtendedApduFormatter(_MaxApduSize.YK4_3)

    def __init__(
        self, delegate: ApduProcessor, formatter: ApduFormatter, scp_state: ScpState
    ):
        self.delegate = delegate
        self.formatter = formatter
        self.state = scp_state

    def send_apdu(self, cla, ins, p1, p2, data, le, encrypt: bool = True):
        cla |= 0x04

        if encrypt:
            logger.log(LOG_LEVEL.TRAFFIC, "Plaintext data: %s", data.hex())
            data = self.state.encrypt(data)

        # Always use extended APDU format for MAC calculation if data is long
        formatter = (
            self.extended_formatter
            if len(data) + 8 > SHORT_APDU_MAX_CHUNK
            else self.formatter
        )
        # Calculate and add MAC to data
        apdu = formatter.format_apdu(cla, ins, p1, p2, data + b"\0" * 8, 0)
        mac = self.state.mac(apdu[:-8])
        data = data + mac

        # Send encrypted and MACed APDU
        resp, sw = self.delegate.send_apdu(cla, ins, p1, p2, data, le)

        # Un-MAC and decrypt, if needed
        if resp:
            resp = self.state.unmac(resp, sw)
            if resp:
                resp = self.state.decrypt(resp)
                logger.log(LOG_LEVEL.TRAFFIC, "Plaintext resp: %s", resp.hex())

        return resp, sw


class SmartCardProtocol:
    """An implementation of the Smart Card protocol."""

    def __init__(
        self,
        smartcard_connection: SmartCardConnection,
        ins_send_remaining: int = INS_SEND_REMAINING,
    ):
        self.connection = smartcard_connection
        self._max_apdu_size = _MaxApduSize.NEO
        self._apdu_format = ApduFormat.SHORT
        self._ins_send_remaining = ins_send_remaining
        self._processor = self._build_base_processor()[0]

    def _build_base_processor(self) -> tuple[ApduProcessor, ApduFormatter]:
        processor: ApduProcessor
        if self._apdu_format == ApduFormat.SHORT:
            # Short APDUs need command chaining
            processor = CommandChainingProcessor(self.connection)
            formatter = processor.formatter
        else:
            formatter = ExtendedApduFormatter(self._max_apdu_size)
            processor = ApduFormatProcessor(self.connection, formatter)

        # Add chained response processor
        processor = ResponseChainingProcessor(processor, self._ins_send_remaining)

        return processor, formatter

    def _reconfigure_processor(self) -> None:
        processor, formatter = self._build_base_processor()

        if isinstance(self._processor, ScpProcessor):
            # Keep the existing SCP state
            processor = ScpProcessor(processor, formatter, self._processor.state)

        self._processor = processor

    @property
    def apdu_format(self) -> ApduFormat:
        warnings.warn(
            "Deprecated: do not read apdu_format.",
            DeprecationWarning,
        )

        return self._apdu_format

    @apdu_format.setter
    def apdu_format(self, value) -> None:
        warnings.warn(
            "Deprecated: do not set apdu_format, instead use configure()",
            DeprecationWarning,
        )
        if value == self._apdu_format:
            return

        self._apdu_format = value
        self._reconfigure_processor()

    def close(self) -> None:
        self.connection.close()

    def enable_touch_workaround(self, version: Version) -> None:
        warnings.warn(
            "Deprecated: use SmartCardProtocol.configure(version) instead.",
            DeprecationWarning,
        )
        self._do_enable_touch_workaround(version)

    def _do_enable_touch_workaround(
        self, version: Version, force_short: bool = False
    ) -> bool:
        if self.connection.transport == TRANSPORT.USB and (
            (4, 2, 0) <= version <= (4, 2, 6)
        ):
            self._max_apdu_size = _MaxApduSize.YK4
            if not force_short:
                self._apdu_format = ApduFormat.EXTENDED
            self._processor = TouchWorkaroundProcessor(self._build_base_processor()[0])
            logger.debug("Touch workaround enabled")
            return True
        return False

    def configure(self, version: Version, force_short: bool = False) -> None:
        """Configure the connection optimally for the given YubiKey version."""
        if self._do_enable_touch_workaround(version, force_short):
            # Devices that require the touch workaround don't support additional options
            return

        if version[0] <= 3:
            # Nothing to do for YubiKey NEO
            return

        if self.connection.transport == TRANSPORT.USB and not force_short:
            self._apdu_format = ApduFormat.EXTENDED
        self._max_apdu_size = (
            _MaxApduSize.YK4_3 if version >= (4, 3) else _MaxApduSize.YK4
        )

        self._reconfigure_processor()

    def send_apdu(
        self,
        cla: int,
        ins: int,
        p1: int,
        p2: int,
        data: bytes = b"",
        le: int = 0,
    ) -> bytes:
        """Send APDU message.

        :param cla: The instruction class.
        :param ins: The instruction code.
        :param p1: The instruction parameter.
        :param p2: The instruction parameter.
        :param data: The command data in bytes.
        :param le: The maximum number of bytes in the data
            field of the response.
        """
        resp, sw = self._processor.send_apdu(cla, ins, p1, p2, data, le)

        if sw != SW.OK:
            raise ApduError(resp, sw)

        return resp

    def select(self, aid: bytes) -> bytes:
        """Perform a SELECT instruction.

        :param aid: The YubiKey application AID value.
        """
        logger.debug(f"Selecting AID: {aid.hex()}")
        # Reset the APDU processor to the base state for the SELECT command
        self._processor = self._build_base_processor()[0]

        try:
            return self.send_apdu(0, INS_SELECT, P1_SELECT, P2_SELECT, aid)
        except ApduError as e:
            if e.sw in (
                SW.FILE_NOT_FOUND,
                SW.APPLET_SELECT_FAILED,
                SW.INVALID_INSTRUCTION,
                SW.WRONG_PARAMETERS_P1P2,
            ):
                raise ApplicationNotAvailableError()
            raise

    def init_scp(self, key_params: ScpKeyParams) -> None:
        # Reset the APDU processor to the base state
        self._processor, formatter = self._build_base_processor()
        try:
            if isinstance(key_params, Scp03KeyParams):
                self._scp03_init(key_params, formatter)
            elif isinstance(key_params, Scp11KeyParams):
                self._scp11_init(key_params, formatter)
            else:
                raise ValueError("Unsupported ScpKeyParams")

            logger.info("SCP initialized")
        except ApduError as e:
            if e.sw == SW.CLASS_NOT_SUPPORTED:
                raise NotSupportedError(
                    "This YubiKey does not support secure messaging"
                )
            if e.sw == SW.REFERENCE_DATA_NOT_FOUND:
                raise ValueError("Incorrect SCP parameters")
            raise
        except BadResponseError:
            raise ValueError("Incorrect SCP parameters")

    def _scp03_init(self, key_params: Scp03KeyParams, formatter: ApduFormatter) -> None:
        logger.debug("Initializing SCP03")
        scp, host_cryptogram = ScpState.scp03_init(self.send_apdu, key_params)
        self._processor = ScpProcessor(self._processor, formatter, scp)

        # Send EXTERNAL AUTHENTICATE
        # P1 = C-DECRYPTION, R-ENCRYPTION, C-MAC, and R-MAC
        self._processor.send_apdu(
            0x84, INS_EXTERNAL_AUTHENTICATE, 0x33, 0, host_cryptogram, 0, encrypt=False
        )

    def _scp11_init(self, key_params: Scp11KeyParams, formatter: ApduFormatter) -> None:
        logger.debug("Initializing SCP11")
        scp = ScpState.scp11_init(self.send_apdu, key_params)
        self._processor = ScpProcessor(self._processor, formatter, scp)



import datetime
import logging
import os

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509.oid import NameOID

logger = logging.getLogger(__name__)

PKG_DIR = os.path.dirname(os.path.abspath(__file__))


def open_file(*relative_path):
    return open(os.path.join(PKG_DIR, "files", *relative_path), "rb")


def generate_self_signed_certificate(
    common_name="Test", valid_from=None, valid_to=None
):
    valid_from = (
        valid_from if valid_from else datetime.datetime.now(datetime.timezone.utc)
    )
    valid_to = valid_to if valid_to else valid_from + datetime.timedelta(days=1)

    private_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
    public_key = private_key.public_key()

    subject = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name)])

    return (
        x509.CertificateBuilder()
        .public_key(public_key)
        .subject_name(subject)
        .issuer_name(subject)  # Same as subject on self-signed certificate.
        .serial_number(x509.random_serial_number())
        .not_valid_before(valid_from)
        .not_valid_after(valid_to)
    ).sign(private_key, hashes.SHA256(), default_backend())



# Copyright (c) 2020 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import subprocess  # nosec
import sys
from time import sleep

from smartcard import System
from smartcard.Exceptions import CardConnectionException, NoCardException
from smartcard.ExclusiveConnectCardConnection import ExclusiveConnectCardConnection
from smartcard.pcsc.PCSCExceptions import ListReadersException

from yubikit.core import PID, TRANSPORT, YUBIKEY
from yubikit.core.fido import SmartCardCtapDevice
from yubikit.core.smartcard import SmartCardConnection
from yubikit.logging import LOG_LEVEL
from yubikit.management import USB_INTERFACE
from yubikit.support import read_info

from ..base import REINSERT_STATUS, CancelledException, YkmanDevice

logger = logging.getLogger(__name__)


YK_READER_NAME = "yubico yubikey"
_YKMAN_NO_EXCLUSIVE = "YKMAN_NO_EXLUSIVE"


# Figure out what the PID should be based on the reader name
def _pid_from_name(name):
    if YK_READER_NAME not in name.lower():
        return None

    interfaces = USB_INTERFACE(0)
    for iface in USB_INTERFACE:
        if iface.name in name:
            interfaces |= iface

    if "U2F" in name:
        interfaces |= USB_INTERFACE.FIDO

    key_type = YUBIKEY.NEO if "NEO" in name else YUBIKEY.YK4
    return PID.of(key_type, interfaces)


def _release(connection):
    if hasattr(connection, "release"):
        connection.release()


class ScardSmartCardConnection(SmartCardConnection):
    def __init__(self, connection):
        connection.connect()
        self.connection = connection

        atr = self.connection.getATR()
        self._transport = (
            TRANSPORT.USB if atr and atr[1] & 0xF0 == 0xF0 else TRANSPORT.NFC
        )

    @property
    def transport(self):
        return self._transport

    def close(self):
        self.connection.disconnect()
        _release(self.connection)

    def send_and_receive(self, apdu):
        """Sends a command APDU and returns the response data and sw"""
        logger.log(LOG_LEVEL.TRAFFIC, "SEND: %s", apdu.hex())
        data, sw1, sw2 = self.connection.transmit(list(apdu))
        logger.log(
            LOG_LEVEL.TRAFFIC, "RECV: %s SW=%02x%02x", bytes(data).hex(), sw1, sw2
        )
        return bytes(data), sw1 << 8 | sw2


class ScardYubiKeyDevice(YkmanDevice):
    """YubiKey Smart card device"""

    def __init__(self, reader):
        # Base transport on reader name: NFC readers will have a different name
        if YK_READER_NAME in reader.name.lower():
            transport = TRANSPORT.USB
        else:
            transport = TRANSPORT.NFC
        super().__init__(transport, reader.name, _pid_from_name(reader.name))
        self.reader = reader

    def supports_connection(self, connection_type):
        if issubclass(SmartCardCtapDevice, connection_type):
            return self.transport == TRANSPORT.NFC
        return issubclass(ScardSmartCardConnection, connection_type)

    def open_connection(self, connection_type):
        assert isinstance(connection_type, type)  # noqa: S101
        if issubclass(ScardSmartCardConnection, connection_type):
            return self._open_smartcard_connection()
        elif issubclass(SmartCardCtapDevice, connection_type):
            return SmartCardCtapDevice(self._open_smartcard_connection())
        return super().open_connection(connection_type)

    def _open_smartcard_connection(self, retry=True) -> SmartCardConnection:
        connection = self.reader.createConnection()
        try:
            # Try an exclusive connection, unless disabled
            if os.environ.get(_YKMAN_NO_EXCLUSIVE) is None:
                excl_connection = ExclusiveConnectCardConnection(connection)
                try:
                    scard_conn = ScardSmartCardConnection(excl_connection)
                    logger.debug("Using exclusive CCID connection")
                    return scard_conn
                except CardConnectionException:
                    logger.info("Failed to get exclusive CCID access")

            # Try a shared connection
            return ScardSmartCardConnection(connection)
        except CardConnectionException:
            _release(connection)
            # Neither connection worked, maybe we need to kill stuff
            if retry and (kill_scdaemon() or kill_yubikey_agent()):
                return self._open_smartcard_connection(False)
            raise
        except (NoCardException, ValueError):
            _release(connection)
            # Handle reclaim timeout
            # TODO: Maybe only on NEO?
            if retry and self.transport == TRANSPORT.USB:
                for _ in range(6):
                    try:
                        sleep(0.5)
                        return self._open_smartcard_connection(False)
                    except (NoCardException, ValueError):
                        continue
            raise

    def _do_reinsert(self, reinsert_cb, event):
        removed = False
        with self.open_connection(SmartCardConnection) as conn:
            info = read_info(conn, self.pid)
        reinsert_cb(REINSERT_STATUS.REMOVE)

        if self.transport == TRANSPORT.NFC:
            while not event.wait(0.5):
                try:
                    conn = self.open_connection(SmartCardConnection)
                    if removed:
                        info2 = read_info(conn, self.pid)
                        conn.close()
                        if info.serial != info2.serial or info.version != info2.version:
                            raise ValueError(
                                "Reinserted YubiKey does not match the original"
                            )
                        sleep(1.0)  # Wait for the device to settle
                        return
                    conn.close()
                except CardConnectionException:
                    pass  # Expected, ignore
                except NoCardException:
                    if not removed:
                        reinsert_cb(REINSERT_STATUS.REINSERT)
                        removed = True

            raise CancelledException()
        else:
            while not event.wait(0.5):
                if not removed:
                    # Wait for the reader to be removed
                    if self.reader not in list_readers():
                        reinsert_cb(REINSERT_STATUS.REINSERT)
                        removed = True
                else:
                    # Wait for the reader to be reinserted
                    for reader in list_readers():
                        if reader.name == self.reader.name:
                            self.reader = reader
                            with self.open_connection(SmartCardConnection) as conn:
                                info2 = read_info(conn, self.pid)
                            if (
                                info.serial != info2.serial
                                or info.version != info2.version
                            ):
                                raise ValueError(
                                    "Reinserted YubiKey does not match the original"
                                )
                            sleep(1.0)  # Wait for the device to settle
                            return
            raise CancelledException()


def kill_scdaemon():
    killed = False
    if sys.platform == "win32":
        # Works for Windows.

        # TODO: Type checking is currently failing on Github Actions
        from typing import TYPE_CHECKING

        if not TYPE_CHECKING:
            from win32api import CloseHandle, OpenProcess, TerminateProcess
            from win32com.client import GetObject

            wmi = GetObject("winmgmts:")
            ps = wmi.InstancesOf("Win32_Process")
            for p in ps:
                if p.Properties_("Name").Value == "scdaemon.exe":
                    pid = p.Properties_("ProcessID").Value
                    handle = OpenProcess(1, False, pid)
                    TerminateProcess(handle, -1)
                    CloseHandle(handle)
                    killed = True
    else:
        # Works for Linux and OS X.
        return_code = subprocess.call(["pkill", "-9", "scdaemon"])  # noqa: S603, S607
        if return_code == 0:
            killed = True
    if killed:
        sleep(0.1)
    return killed


def kill_yubikey_agent():
    killed = False
    if sys.platform != "win32":
        return_code = subprocess.call(["pkill", "-HUP", "yubikey-agent"])  # noqa: S603, S607
        if return_code == 0:
            killed = True
    if killed:
        sleep(0.1)

    return killed


def list_readers():
    try:
        return System.readers()
    except ListReadersException as e:
        # If the PCSC system has restarted the context might be stale, try
        # forcing a new context (This happens on Windows if the last reader is
        # removed):
        try:
            from smartcard.pcsc.PCSCContext import PCSCContext  # type: ignore

            PCSCContext.instance = None
            return System.readers()
        except ImportError:
            # As of pyscard 2.2.2 the PCSCContext singleton has been removed
            raise e


def list_devices(name_filter=None):
    name_filter = YK_READER_NAME if name_filter is None else name_filter
    devices = []
    for reader in list_readers():
        if name_filter.lower() in reader.name.lower():
            devices.append(ScardYubiKeyDevice(reader))
    return devices



import pytest

from yubikit.piv import OBJECT_ID


class TestMisc:
    def setUp(self, ykman_cli):
        ykman_cli("piv", "reset", "-f")

    def test_info(self, ykman_cli):
        output = ykman_cli("piv", "info").output
        assert "PIV version:" in output

    def test_reset(self, ykman_cli):
        output = ykman_cli("piv", "reset", "-f").output
        assert "Reset complete" in output

    def test_export_invalid_certificate_fails(self, ykman_cli, keys):
        ykman_cli(
            "piv",
            "objects",
            "import",
            hex(OBJECT_ID.AUTHENTICATION),
            "-",
            "-m",
            keys.mgmt,
            input="This is not a cert",
        )

        with pytest.raises(SystemExit):
            ykman_cli(
                "piv", "certificates", "export", hex(OBJECT_ID.AUTHENTICATION), "-"
            )

    def test_info_with_invalid_certificate_does_not_crash(self, ykman_cli, keys):
        ykman_cli(
            "piv",
            "objects",
            "import",
            hex(OBJECT_ID.AUTHENTICATION),
            "-",
            "-m",
            keys.mgmt,
            input="This is not a cert",
        )
        ykman_cli("piv", "info")



import os
import tempfile

import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from yubikit.core import NotSupportedError
from yubikit.management import CAPABILITY

from ... import condition


def generate_pem_eccp256_keypair():
    pk = ec.generate_private_key(ec.SECP256R1(), default_backend())
    return (
        pk.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        pk.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ),
    )


def roca(version):
    """Not ROCA affected"""
    return (4, 2, 0) <= version < (4, 3, 5)


def not_roca(version):
    """ROCA affected"""
    return not roca(version)


@pytest.fixture()
def tmp_file():
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    yield tmp.name
    os.remove(tmp.name)


class TestKeyExport:
    @condition.min_version(5, 3)
    def test_from_metadata(self, ykman_cli, keys):
        pair = generate_pem_eccp256_keypair()

        ykman_cli(
            "piv",
            "keys",
            "import",
            "9a",
            "-m",
            keys.mgmt,
            "-",
            input=pair[0],
        )
        exported = ykman_cli("piv", "keys", "export", "9a", "-").stdout_bytes
        assert exported == pair[1]

    @condition.min_version(4, 3)
    def test_from_metadata_or_attestation(self, ykman_cli, keys):
        der = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-F",
            "der",
            "-m",
            keys.mgmt,
            "-",
        ).stdout_bytes
        exported = ykman_cli(
            "piv", "keys", "export", "9a", "-F", "der", "-"
        ).stdout_bytes
        assert der == exported

    def test_from_metadata_or_cert(self, ykman_cli, keys):
        private_key_pem, public_key_pem = generate_pem_eccp256_keypair()
        ykman_cli(
            "piv",
            "keys",
            "import",
            "9a",
            "-m",
            keys.mgmt,
            "-",
            input=private_key_pem,
        )
        ykman_cli(
            "piv",
            "certificates",
            "generate",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            "-s",
            "test",
            input=public_key_pem,
        )

        exported = ykman_cli("piv", "keys", "export", "9a", "-").stdout_bytes

        assert public_key_pem == exported

    @condition.max_version(5, 2, 9)
    def test_from_cert_verify(self, ykman_cli, keys):
        private_key_pem, public_key_pem = generate_pem_eccp256_keypair()
        ykman_cli(
            "piv",
            "keys",
            "import",
            "9a",
            "-m",
            keys.mgmt,
            "-",
            input=private_key_pem,
        )
        ykman_cli(
            "piv",
            "certificates",
            "generate",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            "-s",
            "test",
            input=public_key_pem,
        )
        ykman_cli("piv", "keys", "export", "9a", "--verify", "-P", keys.pin, "-")

    @condition.max_version(5, 2, 9)
    def test_from_cert_verify_fails(self, ykman_cli, keys):
        private_key_pem = generate_pem_eccp256_keypair()[0]
        public_key_pem = generate_pem_eccp256_keypair()[1]
        ykman_cli(
            "piv",
            "keys",
            "import",
            "9a",
            "-m",
            keys.mgmt,
            "-",
            input=private_key_pem,
        )
        ykman_cli(
            "piv",
            "certificates",
            "generate",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            "-s",
            "test",
            input=public_key_pem,
        )
        with pytest.raises(SystemExit):
            ykman_cli("piv", "keys", "export", "9a", "--verify", "-P", keys.pin, "-")


class TestKeyManagement:
    @condition.check(not_roca)
    def test_generate_key_default(self, ykman_cli, keys):
        output = ykman_cli("piv", "keys", "generate", "9a", "-m", keys.mgmt, "-").output
        assert "BEGIN PUBLIC KEY" in output

    @condition.check(roca)
    def test_generate_key_default_cve201715361(self, ykman_cli, keys):
        with pytest.raises(NotSupportedError):
            ykman_cli("piv", "keys", "generate", "9a", "-m", keys.mgmt, "-")

    @condition.check(not_roca)
    @condition.yk4_fips(False)
    def test_generate_key_rsa1024(self, ykman_cli, info, keys):
        if CAPABILITY.PIV in info.fips_capable:
            pytest.skip("RSA1024 not available on FIPS")

        output = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "RSA1024",
            "-m",
            keys.mgmt,
            "-",
        ).output
        assert "BEGIN PUBLIC KEY" in output

    @condition.check(not_roca)
    def test_generate_key_rsa2048(self, ykman_cli, keys):
        output = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "RSA2048",
            "-m",
            keys.mgmt,
            "-",
        ).output
        assert "BEGIN PUBLIC KEY" in output

    @condition.yk4_fips(False)
    @condition.check(roca)
    def test_generate_key_rsa1024_cve201715361(self, ykman_cli, keys):
        with pytest.raises(NotSupportedError):
            ykman_cli(
                "piv",
                "keys",
                "generate",
                "9a",
                "-a",
                "RSA1024",
                "-m",
                keys.mgmt,
                "-",
            )

    @condition.check(roca)
    def test_generate_key_rsa2048_cve201715361(self, ykman_cli, keys):
        with pytest.raises(NotSupportedError):
            ykman_cli(
                "piv",
                "keys",
                "generate",
                "9a",
                "-a",
                "RSA2048",
                "-m",
                keys.mgmt,
                "-",
            )

    def test_generate_key_eccp256(self, ykman_cli, keys):
        output = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
        ).output
        assert "BEGIN PUBLIC KEY" in output

    def test_import_key_eccp256(self, ykman_cli, keys):
        ykman_cli(
            "piv",
            "keys",
            "import",
            "9a",
            "-m",
            keys.mgmt,
            "-",
            input=generate_pem_eccp256_keypair()[0],
        )

    @condition.min_version(4)
    def test_generate_key_eccp384(self, ykman_cli, keys):
        output = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP384",
            "-m",
            keys.mgmt,
            "-",
        ).output
        assert "BEGIN PUBLIC KEY" in output

    @condition.min_version(4)
    def test_generate_key_pin_policy_always(self, ykman_cli, keys):
        output = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "--pin-policy",
            "ALWAYS",
            "-m",
            keys.mgmt,
            "-a",
            "ECCP256",
            "-",
        ).output
        assert "BEGIN PUBLIC KEY" in output

    @condition.min_version(4)
    def test_import_key_pin_policy_always(self, ykman_cli, keys):
        for pin_policy in ["ALWAYS", "always"]:
            ykman_cli(
                "piv",
                "keys",
                "import",
                "9a",
                "--pin-policy",
                pin_policy,
                "-m",
                keys.mgmt,
                "-",
                input=generate_pem_eccp256_keypair()[0],
            )

    @condition.min_version(4)
    def test_generate_key_touch_policy_always(self, ykman_cli, keys):
        output = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "--touch-policy",
            "ALWAYS",
            "-m",
            keys.mgmt,
            "-a",
            "ECCP256",
            "-",
        ).output
        assert "BEGIN PUBLIC KEY" in output

    @condition.min_version(4)
    def test_import_key_touch_policy_always(self, ykman_cli, keys):
        for touch_policy in ["ALWAYS", "always"]:
            ykman_cli(
                "piv",
                "keys",
                "import",
                "9a",
                "--touch-policy",
                touch_policy,
                "-m",
                keys.mgmt,
                "-",
                input=generate_pem_eccp256_keypair()[0],
            )

    @condition.min_version(4, 3)
    def test_attest_key(self, ykman_cli, keys):
        ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
        )
        output = ykman_cli("piv", "keys", "attest", "9a", "-").output
        assert "BEGIN CERTIFICATE" in output

    def _test_generate_csr(self, ykman_cli, keys, tmp_file, algo):
        ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            algo,
            "-m",
            keys.mgmt,
            tmp_file,
        )
        output = ykman_cli(
            "piv",
            "certificates",
            "request",
            "9a",
            tmp_file,
            "-s",
            "test-subject",
            "-P",
            keys.pin,
            "-",
        ).output
        csr = x509.load_pem_x509_csr(output.encode(), default_backend())
        assert csr.is_signature_valid

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_rsa1024(self, ykman_cli, keys, info, tmp_file):
        if CAPABILITY.PIV in info.fips_capable:
            pytest.skip("RSA1024 not available on FIPS")

        self._test_generate_csr(ykman_cli, keys, tmp_file, "RSA1024")

    def test_generate_csr_eccp256(self, ykman_cli, keys, tmp_file):
        self._test_generate_csr(ykman_cli, keys, tmp_file, "ECCP256")

    def test_import_verify_correct_cert_succeeds_with_pin(
        self, ykman_cli, keys, tmp_file
    ):
        # Set up a key in the slot and create a certificate for it
        public_key_pem = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
        ).output

        ykman_cli(
            "piv",
            "certificates",
            "generate",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            "-s",
            "test",
            input=public_key_pem,
        )

        ykman_cli("piv", "certificates", "export", "9a", tmp_file)

        with pytest.raises(SystemExit):
            ykman_cli(
                "piv",
                "certificates",
                "import",
                "--verify",
                "9a",
                tmp_file,
                "-m",
                keys.mgmt,
            )

        ykman_cli(
            "piv",
            "certificates",
            "import",
            "--verify",
            "9a",
            tmp_file,
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
        )
        ykman_cli(
            "piv",
            "certificates",
            "import",
            "--verify",
            "9a",
            tmp_file,
            "-m",
            keys.mgmt,
            input=keys.pin,
        )

    def test_import_verify_wrong_cert_fails(self, ykman_cli, keys):
        # Set up a key in the slot and create a certificate for it
        public_key_pem = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
        ).output

        ykman_cli(
            "piv",
            "certificates",
            "generate",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            "-s",
            "test",
            input=public_key_pem,
        )

        cert_pem = ykman_cli("piv", "certificates", "export", "9a", "-").output

        # Overwrite the key with a new one
        ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
            input=public_key_pem,
        )

        with pytest.raises(SystemExit):
            ykman_cli(
                "piv",
                "certificates",
                "import",
                "--verify",
                "9a",
                "-",
                "-m",
                keys.mgmt,
                "-P",
                keys.pin,
                input=cert_pem,
            )

    def test_import_no_verify_wrong_cert_succeeds(self, ykman_cli, keys):
        # Set up a key in the slot and create a certificate for it
        public_key_pem = ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
        ).output

        ykman_cli(
            "piv",
            "certificates",
            "generate",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            "-s",
            "test",
            input=public_key_pem,
        )

        cert_pem = ykman_cli("piv", "certificates", "export", "9a", "-").output

        # Overwrite the key with a new one
        ykman_cli(
            "piv",
            "keys",
            "generate",
            "9a",
            "-a",
            "ECCP256",
            "-m",
            keys.mgmt,
            "-",
            input=public_key_pem,
        )

        with pytest.raises(SystemExit):
            ykman_cli(
                "piv",
                "certificates",
                "import",
                "--verify",
                "9a",
                "-",
                "-m",
                keys.mgmt,
                "-P",
                keys.pin,
                input=cert_pem,
            )

        ykman_cli(
            "piv",
            "certificates",
            "import",
            "9a",
            "-",
            "-m",
            keys.mgmt,
            "-P",
            keys.pin,
            input=cert_pem,
        )

    @condition.min_version(4, 3)
    def test_export_attestation_certificate(self, ykman_cli):
        output = ykman_cli("piv", "certificates", "export", "f9", "-").output
        assert "BEGIN CERTIFICATE" in output



import pytest

from .. import condition


class TestYkmanInfo:
    def test_ykman_info(self, ykman_cli, info):
        output = ykman_cli("info").output
        assert "Device type:" in output
        if info.serial is not None:
            assert "Serial number:" in output
        assert "Firmware version:" in output

    @condition.yk4_fips(False)
    def test_ykman_info_does_not_report_fips_for_non_fips_device(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("info", "--check-fips")

    @condition.yk4_fips(True)
    def test_ykman_info_reports_fips_status(self, ykman_cli):
        info = ykman_cli("info", "--check-fips").output
        assert "FIPS Approved Mode:" in info
        assert "  FIDO U2F:" in info
        assert "  OATH:" in info
        assert "  OTP:" in info



import pytest

from yubikit.management import CAPABILITY

from .. import condition

DEFAULT_PIN = "123456"
NON_DEFAULT_PIN = "12345679"
NON_DEFAULT_PIN_2 = "12345670"
DEFAULT_ADMIN_PIN = "12345678"
NON_DEFAULT_ADMIN_PIN = "12345670"
NON_DEFAULT_ADMIN_PIN_2 = "12345679"


def old_new_new(old, new):
    return f"{old}\n{new}\n{new}\n"


@pytest.fixture(autouse=True)
@condition.capability(CAPABILITY.OPENPGP)
def preconditions(ykman_cli):
    ykman_cli("openpgp", "reset", "-f")


class TestOpenPGP:
    def test_openpgp_info(self, ykman_cli):
        output = ykman_cli("openpgp", "info").output
        assert "OpenPGP version:" in output

    def test_openpgp_reset(self, ykman_cli):
        output = ykman_cli("openpgp", "reset", "-f").output
        assert "Reset complete" in output


class TestPin:
    def test_change_pin(self, ykman_cli):
        ykman_cli(
            "openpgp", "access", "change-pin", "-P", DEFAULT_PIN, "-n", NON_DEFAULT_PIN
        )
        ykman_cli(
            "openpgp",
            "access",
            "change-pin",
            "-P",
            NON_DEFAULT_PIN,
            "-n",
            NON_DEFAULT_PIN_2,
        )

    def test_change_pin_prompt(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "change-pin",
            input=old_new_new(DEFAULT_PIN, NON_DEFAULT_PIN),
        )
        ykman_cli(
            "openpgp",
            "access",
            "change-pin",
            input=old_new_new(NON_DEFAULT_PIN, NON_DEFAULT_PIN_2),
        )


class TestAdminPin:
    def test_change_admin_pin(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "change-admin-pin",
            "-a",
            DEFAULT_ADMIN_PIN,
            "-n",
            NON_DEFAULT_ADMIN_PIN,
        )
        ykman_cli(
            "openpgp",
            "access",
            "change-admin-pin",
            "-a",
            NON_DEFAULT_ADMIN_PIN,
            "-n",
            NON_DEFAULT_ADMIN_PIN_2,
        )

    def test_change_pin_prompt(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "change-admin-pin",
            input=old_new_new(DEFAULT_ADMIN_PIN, NON_DEFAULT_ADMIN_PIN),
        )
        ykman_cli(
            "openpgp",
            "access",
            "change-admin-pin",
            input=old_new_new(NON_DEFAULT_ADMIN_PIN, NON_DEFAULT_ADMIN_PIN_2),
        )


class TestResetPin:
    def ensure_pin_changed(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "change-pin",
            "-P",
            NON_DEFAULT_PIN,
            "-n",
            NON_DEFAULT_PIN_2,
        )

    def test_set_and_use_reset_code(self, ykman_cli):
        reset_code = "00112233"

        ykman_cli(
            "openpgp",
            "access",
            "change-reset-code",
            "-a",
            DEFAULT_ADMIN_PIN,
            "-r",
            reset_code,
        )

        ykman_cli(
            "openpgp",
            "access",
            "unblock-pin",
            "-r",
            reset_code,
            "-n",
            NON_DEFAULT_PIN,
        )

        self.ensure_pin_changed(ykman_cli)

    def test_set_and_use_reset_code_prompt(self, ykman_cli):
        reset_code = "44332211"

        ykman_cli(
            "openpgp",
            "access",
            "change-reset-code",
            input=old_new_new(DEFAULT_ADMIN_PIN, reset_code),
        )

        ykman_cli(
            "openpgp",
            "access",
            "unblock-pin",
            input=old_new_new(reset_code, NON_DEFAULT_PIN),
        )

        ykman_cli(
            "openpgp",
            "access",
            "change-pin",
            "-P",
            NON_DEFAULT_PIN,
            "-n",
            NON_DEFAULT_PIN_2,
        )

    def test_unblock_pin_with_admin_pin(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "unblock-pin",
            "-a",
            DEFAULT_ADMIN_PIN,
            "-n",
            NON_DEFAULT_PIN,
        )

        self.ensure_pin_changed(ykman_cli)

    def test_unblock_pin_with_admin_pin_prompt(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "unblock-pin",
            "--admin-pin",
            "-",
            input=old_new_new(DEFAULT_ADMIN_PIN, NON_DEFAULT_PIN),
        )

        self.ensure_pin_changed(ykman_cli)


class TestForceSignature:
    def test_set_force_sig(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "set-signature-policy",
            "ALWAYS",
            "-a",
            DEFAULT_ADMIN_PIN,
        )

        output = ykman_cli("openpgp", "info").output
        assert "Always" in output

        ykman_cli(
            "openpgp", "access", "set-signature-policy", "ONCE", "-a", DEFAULT_ADMIN_PIN
        )

        output = ykman_cli("openpgp", "info").output
        assert "Once" in output

    def test_set_force_sig_prompt(self, ykman_cli):
        ykman_cli(
            "openpgp",
            "access",
            "set-signature-policy",
            "ALWAYS",
            input=DEFAULT_ADMIN_PIN,
        )

        output = ykman_cli("openpgp", "info").output
        assert "Always" in output

        ykman_cli(
            "openpgp",
            "access",
            "set-signature-policy",
            "ONCE",
            input=DEFAULT_ADMIN_PIN,
        )

        output = ykman_cli("openpgp", "info").output
        assert "Once" in output



from yubikit.core import TRANSPORT
from yubikit.core.fido import FidoConnection
from yubikit.core.otp import OtpConnection
from yubikit.core.smartcard import SmartCardConnection
from yubikit.support import read_info

from . import condition


def try_connection(device, conn_type):
    with device.open_connection(conn_type) as conn:
        read_info(conn, device.pid)
        return True


@condition.transport(TRANSPORT.USB)
def test_switch_interfaces(device):
    for conn_type in (
        FidoConnection,
        OtpConnection,
        FidoConnection,
        SmartCardConnection,
        OtpConnection,
        SmartCardConnection,
        FidoConnection,
    ):
        if device.pid.supports_connection(conn_type):
            assert try_connection(device, conn_type)



from dataclasses import replace
from typing import cast

import pytest
from yubikit.core import TRANSPORT, YUBIKEY, Tlv
from yubikit.management import (
    CAPABILITY,
    FORM_FACTOR,
    DeviceConfig,
    DeviceInfo,
    Version,
)
from yubikit.support import get_name

DEFAULT_INFO = DeviceInfo(
    config=cast(DeviceConfig, None),
    serial=None,
    version=Version(5, 3, 0),
    form_factor=FORM_FACTOR.USB_A_KEYCHAIN,
    supported_capabilities={TRANSPORT.USB: CAPABILITY(0xFF)},
    is_locked=False,
    is_fips=False,
)


def info(form_factor, **kwargs):
    return replace(DEFAULT_INFO, form_factor=form_factor, **kwargs)


def info_nfc(form_factor, **kwargs):
    return replace(
        DEFAULT_INFO,
        form_factor=form_factor,
        supported_capabilities={
            TRANSPORT.USB: CAPABILITY(0xFF),
            TRANSPORT.NFC: CAPABILITY(0xFF),
        },
        **kwargs,
    )


def test_yk5_formfactors():
    kt = YUBIKEY.YK4
    assert get_name(info(FORM_FACTOR.USB_A_KEYCHAIN), kt) == "YubiKey 5A"
    assert get_name(info_nfc(FORM_FACTOR.USB_A_KEYCHAIN), kt) == "YubiKey 5 NFC"
    assert get_name(info(FORM_FACTOR.USB_A_NANO), kt) == "YubiKey 5 Nano"
    assert get_name(info(FORM_FACTOR.USB_C_KEYCHAIN), kt) == "YubiKey 5C"
    assert get_name(info_nfc(FORM_FACTOR.USB_C_KEYCHAIN), kt) == "YubiKey 5C NFC"
    assert get_name(info(FORM_FACTOR.USB_C_NANO), kt) == "YubiKey 5C Nano"
    assert get_name(info(FORM_FACTOR.USB_C_LIGHTNING), kt) == "YubiKey 5Ci"
    assert (
        get_name(info(FORM_FACTOR.USB_A_BIO), kt)
        == "YubiKey Bio - Multi-protocol Edition"
    )
    assert (
        get_name(info(FORM_FACTOR.USB_C_BIO), kt)
        == "YubiKey C Bio - Multi-protocol Edition"
    )
    assert get_name(info(FORM_FACTOR.UNKNOWN), kt) == "YubiKey 5"
    assert get_name(info_nfc(FORM_FACTOR.UNKNOWN), kt) == "YubiKey 5 NFC"


def fido(device_info):
    device_info.supported_capabilities[TRANSPORT.USB] = (
        CAPABILITY.U2F | CAPABILITY.FIDO2
    )
    if TRANSPORT.NFC in device_info.supported_capabilities:
        device_info.supported_capabilities[TRANSPORT.NFC] = (
            CAPABILITY.U2F | CAPABILITY.FIDO2
        )
    return device_info


def test_yk5_fido():
    kt = YUBIKEY.YK4
    assert (
        get_name(fido(info(FORM_FACTOR.USB_A_BIO)), kt) == "YubiKey Bio - FIDO Edition"
    )
    assert (
        get_name(fido(info(FORM_FACTOR.USB_C_BIO)), kt)
        == "YubiKey C Bio - FIDO Edition"
    )


def fips(device_info):
    device_info.is_fips = True
    return device_info


def test_yk5_fips_formfactors():
    kt = YUBIKEY.YK4
    assert get_name(fips(info(FORM_FACTOR.USB_A_KEYCHAIN)), kt) == "YubiKey 5A FIPS"
    assert (
        get_name(fips(info_nfc(FORM_FACTOR.USB_A_KEYCHAIN)), kt) == "YubiKey 5 NFC FIPS"
    )
    assert get_name(fips(info(FORM_FACTOR.USB_A_NANO)), kt) == "YubiKey 5 Nano FIPS"
    assert get_name(fips(info(FORM_FACTOR.USB_C_KEYCHAIN)), kt) == "YubiKey 5C FIPS"
    assert (
        get_name(fips(info_nfc(FORM_FACTOR.USB_C_KEYCHAIN)), kt)
        == "YubiKey 5C NFC FIPS"
    )
    assert get_name(fips(info(FORM_FACTOR.USB_C_NANO)), kt) == "YubiKey 5C Nano FIPS"
    assert get_name(fips(info(FORM_FACTOR.USB_C_LIGHTNING)), kt) == "YubiKey 5Ci FIPS"
    assert get_name(fips(info(FORM_FACTOR.UNKNOWN)), kt) == "YubiKey 5 FIPS"
    assert get_name(fips(info_nfc(FORM_FACTOR.UNKNOWN)), kt) == "YubiKey 5 NFC FIPS"


def sky(device_info):
    device_info.is_sky = True
    return device_info


def test_sky_formfactors():
    kt = YUBIKEY.YK4
    assert get_name(sky(info(FORM_FACTOR.USB_A_KEYCHAIN)), kt) == "Security Key A"
    assert get_name(sky(info_nfc(FORM_FACTOR.USB_A_KEYCHAIN)), kt) == "Security Key NFC"
    assert get_name(sky(info(FORM_FACTOR.USB_A_NANO)), kt) == "Security Key Nano"
    assert get_name(sky(info(FORM_FACTOR.USB_C_KEYCHAIN)), kt) == "Security Key C"
    assert (
        get_name(sky(info_nfc(FORM_FACTOR.USB_C_KEYCHAIN)), kt) == "Security Key C NFC"
    )
    assert get_name(sky(info(FORM_FACTOR.USB_C_NANO)), kt) == "Security Key C Nano"
    assert get_name(sky(info(FORM_FACTOR.USB_C_LIGHTNING)), kt) == "Security Key Ci"
    assert get_name(sky(info(FORM_FACTOR.UNKNOWN)), kt) == "Security Key"
    assert get_name(sky(info_nfc(FORM_FACTOR.UNKNOWN)), kt) == "Security Key NFC"


def skyep(device_info):
    return replace(device_info, is_sky=True, serial=123456)


def test_sky_enterprise_formfactors():
    kt = YUBIKEY.YK4
    assert (
        get_name(skyep(info(FORM_FACTOR.USB_A_KEYCHAIN)), kt)
        == "Security Key A - Enterprise Edition"
    )
    assert (
        get_name(skyep(info_nfc(FORM_FACTOR.USB_A_KEYCHAIN)), kt)
        == "Security Key NFC - Enterprise Edition"
    )
    assert (
        get_name(skyep(info(FORM_FACTOR.USB_A_NANO)), kt)
        == "Security Key Nano - Enterprise Edition"
    )
    assert (
        get_name(skyep(info(FORM_FACTOR.USB_C_KEYCHAIN)), kt)
        == "Security Key C - Enterprise Edition"
    )
    assert (
        get_name(skyep(info_nfc(FORM_FACTOR.USB_C_KEYCHAIN)), kt)
        == "Security Key C NFC - Enterprise Edition"
    )
    assert (
        get_name(skyep(info(FORM_FACTOR.USB_C_NANO)), kt)
        == "Security Key C Nano - Enterprise Edition"
    )
    assert (
        get_name(skyep(info(FORM_FACTOR.USB_C_LIGHTNING)), kt)
        == "Security Key Ci - Enterprise Edition"
    )
    assert (
        get_name(skyep(info(FORM_FACTOR.UNKNOWN)), kt)
        == "Security Key - Enterprise Edition"
    )
    assert (
        get_name(skyep(info_nfc(FORM_FACTOR.UNKNOWN)), kt)
        == "Security Key NFC - Enterprise Edition"
    )


def epin(device_info):
    return replace(device_info, pin_complexity=True)


def test_enhanced_pin():
    kt = YUBIKEY.YK4
    assert (
        get_name(epin(info(FORM_FACTOR.USB_A_KEYCHAIN)), kt)
        == "YubiKey 5A - Enhanced PIN"
    )
    assert (
        get_name(epin(info_nfc(FORM_FACTOR.USB_A_KEYCHAIN)), kt)
        == "YubiKey 5 NFC - Enhanced PIN"
    )
    assert (
        get_name(epin(info_nfc(FORM_FACTOR.USB_C_KEYCHAIN)), kt)
        == "YubiKey 5C NFC - Enhanced PIN"
    )


DEVICE_INFO_TESTDATA = [
    (
        "YubiKey 5C NFC - Enhanced PIN",
        "0102033f0302033f0204020b885b04010305030507040602000007010f0801000d02033f0e02033f0a01000f010020030000002103000000100101110400000000120100130c3738434c5546583530303050140200001502000016010117010018020000",
    ),
    (
        "YubiKey 5C NFC FIPS",
        "0102033d0302033d020401ea505104018305030507040602000007010f0801000d02033d0e02033d0a01000f010020030000002103000000100101110400000000120100130c3738434c55465835303030501402001f1502001f16010117010018020000",
    ),
    (
        "YubiKey 5 Nano",
        "0102033f0302033e020401c234c404010205030507010602000007010f0801000a01000f010020030000002103000000100101110400000000120100130c3738434c55465835303030501402000015020008160100170100",
    ),
    (
        "YubiKey Bio - FIDO Edition",
        "0102020203020202020401afba5c04010605030507020602000007010f0801000a01000f010020030005012103030001100101110400000000120100130c3738434c5546583530303050140200001502000016010017010018020000",
    ),
    (
        "YubiKey C Bio - Multi-protocol Edition",
        "0102061203020612020401afcfff04010705030507020602000007010f0801000a01000f010020030005012103030001100101110400000000120100130c3738434c5546583530303050140200001502000016010017010018020200",
    ),
    (
        "Security Key C NFC - Enterprise Edition",
        "01020202030202020204017b306c04014305030504030602000007010f0801000d0202060e0202060a01000f0100",
    ),
    (
        "YubiKey 5Ci",
        "0102023f0302023f020400a0392204010505030502030602000007010f0801000a01000f010c",
    ),
    (
        "YubiKey C Bio - FIDO Edition",
        "01021206030212060204023251e904010705030000010602000007010f0801000a01000f010020030100002103030001100101110400000000120100130c3738434c5546583530303050140200001502000016010017010018020000190e0103050800020101030400000003",
    ),
]


@pytest.mark.parametrize("expected_name, data", DEVICE_INFO_TESTDATA)
def test_device_info(expected_name, data):
    kt = YUBIKEY.YK4
    tlvs = Tlv.parse_dict(bytes.fromhex(data))
    info = DeviceInfo.parse_tlvs(tlvs, Version(5, 3, 0))
    assert get_name(info, kt) == expected_name



# Copyright (c) 2023 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import abc
import logging
import os
import struct
from dataclasses import dataclass, field
from enum import IntEnum, unique
from typing import Callable, NamedTuple, Sequence

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import cmac, constant_time, hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.x963kdf import X963KDF

from .. import BadResponseError, Tlv

logger = logging.getLogger(__name__)


INS_INITIALIZE_UPDATE = 0x50
INS_EXTERNAL_AUTHENTICATE = 0x82
INS_INTERNAL_AUTHENTICATE = 0x88
INS_PERFORM_SECURITY_OPERATION = 0x2A

_DEFAULT_KEY = b"\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f"

_KEY_ENC = 0x04
_KEY_MAC = 0x06
_KEY_RMAC = 0x07
_CARD_CRYPTOGRAM = 0x00
_HOST_CRYPTOGRAM = 0x01


def _derive(key: bytes, t: int, context: bytes, L: int = 0x80) -> bytes:
    # this only supports aes128
    if L != 0x80 and L != 0x40:
        raise ValueError("L must be 0x40 or 0x80")

    i = b"\0" * 11 + struct.pack("!BBHB", t, 0, L, 1) + context

    c = cmac.CMAC(algorithms.AES(key), backend=default_backend())
    c.update(i)
    return c.finalize()[: L // 8]


def _calculate_mac(key: bytes, chain: bytes, message: bytes) -> tuple[bytes, bytes]:
    c = cmac.CMAC(algorithms.AES(key), backend=default_backend())
    c.update(chain)
    c.update(message)
    chain = c.finalize()
    return chain, chain[:8]


def _init_cipher(key: bytes, counter: int, response=False) -> Cipher:
    encryptor = Cipher(
        algorithms.AES(key),
        modes.ECB(),  # noqa: S305
        backend=default_backend(),
    ).encryptor()
    iv_data = (b"\x80" if response else b"\x00") + int.to_bytes(counter, 15, "big")
    iv = encryptor.update(iv_data) + encryptor.finalize()
    return Cipher(
        algorithms.AES(key),
        modes.CBC(iv),
        backend=default_backend(),
    )


class SessionKeys(NamedTuple):
    """SCP Session Keys."""

    key_senc: bytes
    key_smac: bytes
    key_srmac: bytes
    key_dek: bytes | None = None


class StaticKeys(NamedTuple):
    """SCP03 Static Keys."""

    key_enc: bytes
    key_mac: bytes
    key_dek: bytes | None = None

    @classmethod
    def default(cls) -> StaticKeys:
        return cls(_DEFAULT_KEY, _DEFAULT_KEY, _DEFAULT_KEY)

    def derive(self, context: bytes) -> SessionKeys:
        return SessionKeys(
            _derive(self.key_enc, _KEY_ENC, context),
            _derive(self.key_mac, _KEY_MAC, context),
            _derive(self.key_mac, _KEY_RMAC, context),
            self.key_dek,
        )


@unique
class ScpKid(IntEnum):
    SCP03 = 0x1
    SCP11a = 0x11
    SCP11b = 0x13
    SCP11c = 0x15


class KeyRef(bytes):
    @property
    def kid(self) -> int:
        return self[0]

    @property
    def kvn(self) -> int:
        return self[1]

    def __new__(cls, kid_or_data: int | bytes, kvn: int | None = None):
        """This allows creation by passing either binary data, or kid and kvn."""
        if isinstance(kid_or_data, int):  # kid and kvn
            if kvn is None:
                raise ValueError("Missing kvn")
            data = bytes([kid_or_data, kvn])
        else:  # Binary id and version
            if kvn is not None:
                raise ValueError("kvn can only be provided if kid_or_data is a kid")
            data = kid_or_data

        return super(KeyRef, cls).__new__(cls, data)

    def __init__(self, kid_or_data: int | bytes, kvn: int | None = None):
        if len(self) != 2:
            raise ValueError("Incorrect length")

    def __repr__(self):
        return f"KeyRef(kid=0x{self.kid:02x}, kvn=0x{self.kvn:02x})"

    def __str__(self):
        return repr(self)


@dataclass(frozen=True)
class ScpKeyParams(abc.ABC):
    ref: KeyRef


@dataclass(frozen=True)
class Scp03KeyParams(ScpKeyParams):
    ref: KeyRef = KeyRef(ScpKid.SCP03, 0)
    keys: StaticKeys = StaticKeys.default()


@dataclass(frozen=True)
class Scp11KeyParams(ScpKeyParams):
    pk_sd_ecka: ec.EllipticCurvePublicKey
    # For SCP11 a/c we need an OCE key, with its trust chain
    oce_ref: KeyRef | None = None
    sk_oce_ecka: ec.EllipticCurvePrivateKey | None = None
    # Certificate chain for sk_oce_ecka, leaf-last order
    certificates: Sequence[x509.Certificate] = field(default_factory=list)


SendApdu = Callable[[int, int, int, int, bytes], bytes]


class ScpState:
    def __init__(
        self,
        session_keys: SessionKeys,
        mac_chain: bytes = b"\0" * 16,
        enc_counter: int = 1,
    ):
        self._keys = session_keys
        self._mac_chain = mac_chain
        self._enc_counter = enc_counter

    def encrypt(self, data: bytes) -> bytes:
        # Pad the data
        msg = data
        padlen = 15 - len(msg) % 16
        msg += b"\x80"
        msg = msg.ljust(len(msg) + padlen, b"\0")

        # Encrypt
        cipher = _init_cipher(self._keys.key_senc, self._enc_counter)
        encryptor = cipher.encryptor()
        encrypted = encryptor.update(msg) + encryptor.finalize()
        self._enc_counter += 1
        return encrypted

    def mac(self, data: bytes) -> bytes:
        next_mac_chain, mac = _calculate_mac(self._keys.key_smac, self._mac_chain, data)
        self._mac_chain = next_mac_chain
        return mac

    def unmac(self, data: bytes, sw: int) -> bytes:
        msg, mac = data[:-8], data[-8:]
        rmac = _calculate_mac(
            self._keys.key_srmac, self._mac_chain, msg + struct.pack("!H", sw)
        )[1]
        if not constant_time.bytes_eq(mac, rmac):
            raise BadResponseError("Wrong MAC")
        return msg

    def decrypt(self, encrypted: bytes) -> bytes:
        # Decrypt
        cipher = _init_cipher(self._keys.key_senc, self._enc_counter - 1, True)
        decryptor = cipher.decryptor()
        decrypted = decryptor.update(encrypted) + decryptor.finalize()

        # Unpad
        unpadded = decrypted.rstrip(b"\x00")
        if unpadded[-1] != 0x80:
            raise BadResponseError("Wrong padding")
        unpadded = unpadded[:-1]

        return unpadded

    @classmethod
    def scp03_init(
        cls,
        send_apdu: SendApdu,
        key_params: Scp03KeyParams,
        *,
        host_challenge: bytes | None = None,
    ) -> tuple[ScpState, bytes]:
        logger.debug("Initializing SCP03 handshake")
        host_challenge = host_challenge or os.urandom(8)
        resp = send_apdu(
            0x80, INS_INITIALIZE_UPDATE, key_params.ref.kvn, 0x00, host_challenge
        )

        diversification_data = resp[:10]  # noqa: F841
        key_info = resp[10:13]  # noqa: F841
        card_challenge = resp[13:21]
        card_cryptogram = resp[21:29]

        context = host_challenge + card_challenge
        session_keys = key_params.keys.derive(context)

        gen_card_crypto = _derive(
            session_keys.key_smac, _CARD_CRYPTOGRAM, context, 0x40
        )
        if not constant_time.bytes_eq(gen_card_crypto, card_cryptogram):
            # This means wrong keys
            raise BadResponseError("Wrong SCP03 key set")

        host_cryptogram = _derive(
            session_keys.key_smac, _HOST_CRYPTOGRAM, context, 0x40
        )

        return cls(session_keys), host_cryptogram

    @classmethod
    def scp11_init(
        cls,
        send_apdu: SendApdu,
        key_params: Scp11KeyParams,
    ) -> ScpState:
        kid = ScpKid(key_params.ref.kid)
        logger.debug(f"Initializing {kid.name} handshake")

        # GPC v2.3 Amendment F (SCP11) v1.4 §7.1.1
        if kid == ScpKid.SCP11a:
            params = 0b01
        elif kid == ScpKid.SCP11b:
            params = 0b00
        elif kid == ScpKid.SCP11c:
            params = 0b11
        else:
            raise ValueError("Invalid SCP KID")

        if kid in (ScpKid.SCP11a, ScpKid.SCP11c):
            # GPC v2.3 Amendment F (SCP11) v1.4 §7.5
            assert key_params.sk_oce_ecka  # noqa: S101
            n = len(key_params.certificates) - 1
            assert n >= 0  # noqa: S101
            oce_ref = key_params.oce_ref or KeyRef(0, 0)
            logger.debug("Sending certificate chain")
            for i, cert in enumerate(key_params.certificates):
                p2 = oce_ref.kid | (0x80 if i < n else 0)
                data = cert.public_bytes(serialization.Encoding.DER)
                logger.debug(f"Sending cert: {cert.subject}")
                send_apdu(0x80, INS_PERFORM_SECURITY_OPERATION, oce_ref.kvn, p2, data)

        key_usage = bytes(
            [0x3C]
        )  # AUTHENTICATED | C_MAC | C_DECRYPTION | R_MAC | R_ENCRYPTION
        key_type = bytes([0x88])  # AES
        key_len = bytes([16])  # 128-bit

        # Host ephemeral key
        esk_oce_ecka = ec.generate_private_key(key_params.pk_sd_ecka.curve)
        epk_oce_ecka = esk_oce_ecka.public_key()

        # GPC v2.3 Amendment F (SCP11) v1.4 §7.6.2.3
        data = Tlv(
            0xA6,
            Tlv(0x90, bytes([0x11, params]))
            + Tlv(0x95, key_usage)
            + Tlv(0x80, key_type)
            + Tlv(0x81, key_len),
        ) + Tlv(
            0x5F49,
            epk_oce_ecka.public_bytes(
                serialization.Encoding.X962,
                serialization.PublicFormat.UncompressedPoint,
            ),
        )

        # Static host key (SCP11a/c), or ephemeral key again (SCP11b)
        sk_oce_ecka = key_params.sk_oce_ecka or esk_oce_ecka

        logger.debug("Performing key agreement")
        ins = (
            INS_INTERNAL_AUTHENTICATE
            if key_params.ref.kid == ScpKid.SCP11b
            else INS_EXTERNAL_AUTHENTICATE
        )
        resp = send_apdu(0x80, ins, key_params.ref.kvn, key_params.ref.kid, data)

        epk_sd_ecka_tlv, resp = Tlv.parse_from(resp)
        epk_sd_ecka = Tlv.unpack(0x5F49, epk_sd_ecka_tlv)
        receipt = Tlv.unpack(0x86, resp)

        # GPC v2.3 Amendment F (SCP11) v1.3 §3.1.2 Key Derivation
        key_agreement_data = data + epk_sd_ecka_tlv
        sharedinfo = key_usage + key_type + key_len
        keybytes = X963KDF(hashes.SHA256(), 5 * key_len[0], sharedinfo).derive(
            esk_oce_ecka.exchange(
                ec.ECDH(),
                ec.EllipticCurvePublicKey.from_encoded_point(
                    sk_oce_ecka.curve, epk_sd_ecka
                ),
            )
            + sk_oce_ecka.exchange(ec.ECDH(), key_params.pk_sd_ecka)
        )

        # 5 keys were derived, one for verification of receipt
        ln = key_len[0]
        keys = [keybytes[i : i + ln] for i in range(0, ln * 5, ln)]
        c = cmac.CMAC(algorithms.AES(keys.pop(0)))
        c.update(key_agreement_data)
        c.verify(receipt)
        # The 4 remaining keys are session keys
        session_keys = SessionKeys(*keys)

        return cls(session_keys, receipt)



import os

import pytest
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from yubikit.core import TRANSPORT
from yubikit.core.smartcard import ApduError
from yubikit.hsmauth import (
    INITIAL_RETRY_COUNTER,
    Credential,
    HsmAuthSession,
    InvalidPinError,
)
from yubikit.management import CAPABILITY

from . import condition

DEFAULT_MANAGEMENT_KEY = bytes.fromhex("00000000000000000000000000000000")
NON_DEFAULT_MANAGEMENT_KEY = bytes.fromhex("11111111111111111111111111111112")


@pytest.fixture
@condition.capability(CAPABILITY.HSMAUTH)
@condition.min_version(5, 4, 3)
def session(ccid_connection, transport, info, scp_params):
    if transport == TRANSPORT.NFC and CAPABILITY.HSMAUTH in info.fips_capable:
        hsmauth = HsmAuthSession(ccid_connection, scp_params)
    else:
        hsmauth = HsmAuthSession(ccid_connection)
    hsmauth.reset()
    yield hsmauth


@pytest.fixture
def management_key(session, info):
    if CAPABILITY.HSMAUTH in info.fips_capable:
        key = bytes.fromhex("00000000000000000000000000000001")
        session.put_management_key(DEFAULT_MANAGEMENT_KEY, key)

        yield key
    else:
        yield DEFAULT_MANAGEMENT_KEY


def import_key_derived(
    session,
    management_key,
    credential_password="12345679",
    derivation_password="p4ssw0rd",
) -> Credential:
    credential = session.put_credential_derived(
        management_key,
        "Test PUT credential symmetric (derived)",
        derivation_password,
        credential_password,
    )

    return credential


def import_key_symmetric(
    session, management_key, key_enc, key_mac, credential_password="12345679"
) -> Credential:
    credential = session.put_credential_symmetric(
        management_key,
        "Test PUT credential symmetric",
        key_enc,
        key_mac,
        credential_password,
    )

    return credential


def import_key_asymmetric(
    session, management_key, private_key, credential_password="12345679"
) -> Credential:
    credential = session.put_credential_asymmetric(
        management_key,
        "Test PUT credential asymmetric",
        private_key,
        credential_password,
    )

    return credential


def generate_key_asymmetric(
    session, management_key, credential_password="12345679"
) -> Credential:
    credential = session.generate_credential_asymmetric(
        management_key,
        "Test GENERATE credential asymmetric",
        credential_password,
    )

    return credential


class TestCredentialManagement:
    def check_credential_in_list(self, session, credential: Credential):
        credentials = session.list_credentials()

        assert credential in credentials
        credential_retrieved = next(cred for cred in credentials if cred == credential)
        assert credential_retrieved.label == credential.label
        assert credential_retrieved.touch_required == credential.touch_required
        assert credential_retrieved.algorithm == credential.algorithm
        assert credential_retrieved.counter == INITIAL_RETRY_COUNTER

    def verify_credential_password(
        self, session, credential_password: str, credential: Credential
    ):
        context = b"g\xfc\xf1\xfe\xb5\xf1\xd8\x83\xedv=\xbfI0\x90\xbb"

        # Try to calculate session keys using wrong credential password
        with pytest.raises(InvalidPinError):
            session.calculate_session_keys_symmetric(
                label=credential.label,
                context=context,
                credential_password="wrongvalue",
            )

        # Try to calculate session keys using correct credential password
        session.calculate_session_keys_symmetric(
            label=credential.label,
            context=context,
            credential_password=credential_password,
        )

    def test_import_credential_symmetric_wrong_management_key(
        self, session, management_key
    ):
        with pytest.raises(InvalidPinError):
            import_key_derived(session, NON_DEFAULT_MANAGEMENT_KEY)

    def test_import_credential_symmetric_wrong_key_length(
        self, session, management_key
    ):
        with pytest.raises(ValueError):
            import_key_symmetric(
                session, management_key, os.urandom(24), os.urandom(24)
            )

    def test_import_credential_symmetric_exists(self, session, management_key):
        import_key_derived(session, management_key)
        with pytest.raises(ApduError):
            import_key_derived(session, management_key)

    def test_import_credential_symmetric_works(self, session, management_key):
        credential = import_key_derived(session, management_key, "12345679")

        self.verify_credential_password(session, "12345679", credential)
        self.check_credential_in_list(session, credential)

        session.delete_credential(management_key, credential.label)

    @condition.min_version(5, 6)
    def test_import_credential_asymmetric_unsupported_key(
        self, session, management_key
    ):
        private_key = ec.generate_private_key(
            ec.SECP224R1(), backend=default_backend()
        )  # curve secp224r1 is not supported

        with pytest.raises(ValueError):
            import_key_asymmetric(session, management_key, private_key)

    @condition.min_version(5, 6)
    def test_import_credential_asymmetric_works(self, session, management_key):
        private_key = ec.generate_private_key(ec.SECP256R1(), backend=default_backend())
        credential = import_key_asymmetric(session, management_key, private_key)

        public_key = private_key.public_key()
        assert public_key.public_bytes(
            encoding=Encoding.PEM, format=PublicFormat.SubjectPublicKeyInfo
        ) == session.get_public_key(credential.label).public_bytes(
            encoding=Encoding.PEM, format=PublicFormat.SubjectPublicKeyInfo
        )

        self.check_credential_in_list(session, credential)
        session.delete_credential(management_key, credential.label)

    @condition.min_version(5, 6)
    def test_generate_credential_asymmetric_works(self, session, management_key):
        credential = generate_key_asymmetric(session, management_key)

        self.check_credential_in_list(session, credential)

        public_key = session.get_public_key(credential.label)

        assert isinstance(public_key, ec.EllipticCurvePublicKey)
        assert isinstance(public_key.curve, ec.SECP256R1)

        session.delete_credential(management_key, credential.label)

    @condition.min_version(5, 6)
    def test_export_public_key_symmetric_credential(self, session, management_key):
        credential = import_key_derived(session, management_key)

        with pytest.raises(ApduError):
            session.get_public_key(credential.label)

        session.delete_credential(management_key, credential.label)

    def test_delete_credential_wrong_management_key(self, session, management_key):
        credential = import_key_derived(session, management_key)

        with pytest.raises(InvalidPinError):
            session.delete_credential(NON_DEFAULT_MANAGEMENT_KEY, credential.label)

    def test_delete_credential_non_existing(self, session, management_key):
        with pytest.raises(ApduError):
            session.delete_credential(management_key, "Default key")

    def test_delete_credential_works(self, session, management_key):
        credential = import_key_derived(session, management_key)

        session.delete_credential(management_key, credential.label)
        credentials = session.list_credentials()
        assert len(credentials) == 0

    @condition.min_version(5, 8)
    def test_change_credential_password(self, session, management_key):
        credential_password = "987654321"
        new_credential_password = "123456789"
        credential = import_key_derived(session, management_key, credential_password)

        # Try to change credential password using wrong old password
        with pytest.raises(InvalidPinError):
            session.change_credential_password(
                credential.label,
                "wrongvalue",
                new_credential_password,
            )

        session.change_credential_password(
            credential.label,
            credential_password,
            new_credential_password,
        )

        self.verify_credential_password(session, new_credential_password, credential)

        session.delete_credential(management_key, credential.label)

    @condition.min_version(5, 8)
    def test_change_credential_password_admin(self, session, management_key):
        new_credential_password = "123456789"
        credential = import_key_derived(session, management_key)

        # Try to change credential password using wrong management key
        with pytest.raises(InvalidPinError):
            session.change_credential_password_admin(
                credential.label,
                NON_DEFAULT_MANAGEMENT_KEY,
                new_credential_password,
            )

        session.change_credential_password_admin(
            credential.label,
            management_key,
            new_credential_password,
        )

        self.verify_credential_password(session, new_credential_password, credential)

        session.delete_credential(management_key, credential.label)


class TestAccess:
    def test_change_management_key(self, session, management_key):
        session.put_management_key(management_key, NON_DEFAULT_MANAGEMENT_KEY)

        # Can't import key with old management key
        with pytest.raises(InvalidPinError):
            import_key_derived(session, management_key)

        import_key_derived(session, NON_DEFAULT_MANAGEMENT_KEY)

    def test_management_key_retries(self, session, management_key):
        initial_retries = session.get_management_key_retries()
        assert initial_retries == 8

        with pytest.raises(InvalidPinError):
            import_key_derived(session, NON_DEFAULT_MANAGEMENT_KEY)

        post_retries = session.get_management_key_retries()
        assert post_retries == 7


class TestSessionKeys:
    def test_calculate_session_keys_symmetric(self, session, management_key):
        credential_password = "a password"
        credential = import_key_derived(
            session,
            management_key,
            credential_password=credential_password,
            derivation_password="pwd",
        )

        # Example context and session keys
        context = b"g\xfc\xf1\xfe\xb5\xf1\xd8\x83\xedv=\xbfI0\x90\xbb"
        key_senc = b"\xb0o\x1a\xc9\x87\x91.\xbe\xdc\x1b\xf0\xe0*k]\x85"
        key_smac = b"\xea\xd6\xc3\xa5\x96\xea\x86u\xbf1\xd3I\xab\xb5,t"
        key_srmac = b"\xc2\xc6\x1e\x96\xab,X\xe9\x83z\xd0\xe7\xd0n\xe9\x0c"

        session_keys = session.calculate_session_keys_symmetric(
            label=credential.label,
            context=context,
            credential_password=credential_password,
        )

        assert key_senc == session_keys.key_senc
        assert key_smac == session_keys.key_smac
        assert key_srmac == session_keys.key_srmac


class TestHostChallenge:
    @condition.min_version(5, 6)
    def test_get_challenge_symmetric(self, session, management_key):
        credential = import_key_derived(session, management_key)

        challenge1 = session.get_challenge(credential.label)
        challenge2 = session.get_challenge(credential.label)
        assert len(challenge1) == 8
        assert len(challenge2) == 8
        assert challenge1 != challenge2

        session.delete_credential(management_key, credential.label)

    @condition.min_version(5, 6)
    def test_get_challenge_asymmetric(self, session, management_key):
        credential_password = "12345679"
        credential = generate_key_asymmetric(
            session, management_key, credential_password
        )

        challenge1 = session.get_challenge(credential.label, credential_password)
        challenge2 = session.get_challenge(credential.label, credential_password)

        assert len(challenge1) == 65
        assert len(challenge2) == 65
        assert challenge1 != challenge2

        session.delete_credential(management_key, credential.label)



import os

import pytest
from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from yubikit.core import TRANSPORT, Tlv
from yubikit.core.smartcard import ApduError
from yubikit.core.smartcard.scp import (
    KeyRef,
    Scp03KeyParams,
    Scp11KeyParams,
    ScpKid,
    StaticKeys,
)
from yubikit.securitydomain import SecurityDomainSession

from ..util import open_file
from . import condition


@pytest.fixture
@condition.min_version(5, 7, 2)
def session(ccid_connection):
    sd = SecurityDomainSession(ccid_connection)
    # Check for the default keyset and only reset if not present, to save time
    default_key = KeyRef(0x1, 0xFF)
    if default_key not in sd.get_key_information():
        sd.reset()
    return sd


def test_card_recognition_data(session):
    data = session.get_card_recognition_data()
    tlvs = Tlv.parse_list(data)
    tags = [x.tag for x in tlvs]
    assert tags[:5] == [0x06, 0x60, 0x63, 0x64, 0x64]


def _verify_auth(sd):
    ref = KeyRef(0x13, 0x7F)
    sd.generate_ec_key(ref)
    sd.delete_key(ref.kid, ref.kvn)


class TestScp03:
    @pytest.fixture(autouse=True)
    def preconditions(self, info, transport):
        if info.is_fips and transport != TRANSPORT.USB:
            pytest.skip("SCP management on YK FIPS over NFC")

    def test_ok(self, session):
        session.authenticate(Scp03KeyParams())
        _verify_auth(session)

    def test_wrong_key(self, session):
        with pytest.raises(ValueError):
            session.authenticate(Scp03KeyParams(keys=StaticKeys(*((b"\1" * 16,) * 3))))

        with pytest.raises(ApduError):
            _verify_auth(session)

    def test_change_key(self, ccid_connection, session):
        session.authenticate(Scp03KeyParams())
        ref = KeyRef(0x1, 0x2)
        keys = StaticKeys(os.urandom(16), os.urandom(16), os.urandom(16))
        session.put_key(ref, keys)

        # Test new key
        session = SecurityDomainSession(ccid_connection)
        session.authenticate(Scp03KeyParams(keys=keys))
        _verify_auth(session)

        # Verify default key is removed
        session = SecurityDomainSession(ccid_connection)
        with pytest.raises(ValueError):
            session.authenticate(Scp03KeyParams())


def _load_scp11_keys(session, kid, kvn):
    sd_ref = KeyRef(kid, kvn)
    oce_ref = KeyRef(0x10, kvn)

    pub_key = session.generate_ec_key(sd_ref)
    with open_file("scp/cert.ca-kloc.ecdsa.pem") as f:
        ca = x509.load_pem_x509_certificate(f.read())
    session.put_key(oce_ref, ca.public_key())
    session.store_ca_issuer(
        oce_ref,
        ca.extensions.get_extension_for_class(x509.SubjectKeyIdentifier).value.digest,
    )

    with open_file("scp/cert.ka-kloc.ecdsa.pem") as f:
        ka = x509.load_pem_x509_certificate(f.read())

    with open_file("scp/cert.oce.ecka.pem") as f:
        ecka = x509.load_pem_x509_certificate(f.read())

    with open_file("scp/sk.oce.ecka.pem") as f:
        sk = serialization.load_pem_private_key(f.read(), None)

    return Scp11KeyParams(
        sd_ref,
        pub_key,
        oce_ref,
        sk,
        [ka, ecka],
    )


class TestScp11:
    @pytest.fixture(autouse=True)
    def preconditions(self, info, transport):
        if info.is_fips and transport != TRANSPORT.USB:
            pytest.skip("SCP management on YK FIPS over NFC")

    def test_scp11b_ok(self, session):
        ref = KeyRef(0x13, 0x1)
        chain = session.get_certificate_bundle(ref)
        session.authenticate(Scp11KeyParams(ref, chain[-1].public_key()))
        with pytest.raises(ApduError):
            _verify_auth(session)

    def test_scp11b_wrong_pubkey(self, session):
        ref = KeyRef(0x13, 0x1)
        chain = session.get_certificate_bundle(ref)
        with pytest.raises(InvalidSignature):
            # Using the public key from the intermediate cert instead of the leaf
            session.authenticate(Scp11KeyParams(ref, chain[0].public_key()))

    def test_scp11b_import(self, ccid_connection, session):
        session.authenticate(Scp03KeyParams())

        ref = KeyRef(0x13, 0x2)
        sk = ec.generate_private_key(ec.SECP256R1(), default_backend())
        session.put_key(ref, sk)

        params = Scp11KeyParams(ref, sk.public_key())

        # Authenticate
        session = SecurityDomainSession(ccid_connection)
        session.authenticate(params)

    def test_scp11a_ok(self, ccid_connection, session):
        session.authenticate(Scp03KeyParams())
        kvn = 0x3
        params = _load_scp11_keys(session, ScpKid.SCP11a, kvn)
        # Authenticate
        session = SecurityDomainSession(ccid_connection)
        session.authenticate(params)

        # Verify by deleting keys
        session.delete_key(kvn=kvn)

    def test_scp11a_allowlist(self, ccid_connection, session):
        session.authenticate(Scp03KeyParams())
        kvn = 0x3
        params = _load_scp11_keys(session, ScpKid.SCP11a, kvn)
        serials = [c.serial_number for c in params.certificates]
        session.store_allowlist(params.oce_ref, serials)

        # Authenticate
        session = SecurityDomainSession(ccid_connection)
        session.authenticate(params)

        # Verify by deleting keys
        session.delete_key(kvn=kvn)

    def test_scp11a_allowlist_blocked(self, ccid_connection, session):
        session.authenticate(Scp03KeyParams())

        # Replace default SCP03 keys
        ref = KeyRef(0x1, 0x2)
        keys = StaticKeys(os.urandom(16), os.urandom(16), os.urandom(16))
        session.put_key(ref, keys)
        session.delete_key(kid=ScpKid.SCP11b)

        kvn = 0x3
        params = _load_scp11_keys(session, ScpKid.SCP11a, kvn)
        serials = [1, 2, 3, 4, 5]
        session.store_allowlist(params.oce_ref, serials)

        # Fail authentication
        session = SecurityDomainSession(ccid_connection)
        with pytest.raises(ApduError):
            session.authenticate(params)

        # Remove allowlist
        session.authenticate(Scp03KeyParams(keys=keys))
        session.store_allowlist(params.oce_ref, [])

        # Authenticate
        session = SecurityDomainSession(ccid_connection)
        session.authenticate(params)

    def test_scp11c_ok(self, ccid_connection, session):
        session.authenticate(Scp03KeyParams())
        kvn = 0x3
        params = _load_scp11_keys(session, ScpKid.SCP11c, kvn)
        # Authenticate
        session = SecurityDomainSession(ccid_connection)
        session.authenticate(params)

        # Verify not authenticated
        with pytest.raises(ApduError):
            session.delete_key(kvn=kvn)
